
------------------------------------------------------------------------
--
-- AbsorbsViewer
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local AbsorbsMonitor = _G.AbsorbsMonitor;

AbsorbsViewer = LibStub("AceAddon-3.0"):NewAddon("AbsorbsViewer", "AceConsole-3.0", "AceEvent-3.0");

local AceConfig = LibStub("AceConfig-3.0");
local AceConfigDialog = LibStub("AceConfigDialog-3.0");
local AceDB = LibStub("AceDB-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("AbsorbsViewer");

local AbsorbEffects;
local SpellNames;

local playerGUID;

local baseFrame = nil;
local titleFrame = nil;
local framePool_Units = {};
local framePool_Effects = {};
local activeUnitsMap = {};
local activeUnitsList = {};
local lowestUnitFrame = nil;

local isHidden;


-- #################################################################################################


local function GetUnitClassColorRGB(name)
	local _, class = UnitClass(name);

	if _G.RAID_CLASS_COLORS and _G.RAID_CLASS_COLORS[class] then
		return _G.RAID_CLASS_COLORS[class].r, _G.RAID_CLASS_COLORS[class].g, _G.RAID_CLASS_COLORS[class].b;
	else
		return 0.63, 0.63, 0.63;	
	end
end

local function OnDragStart(self)
	if IsAltKeyDown() then
		self.isMoving = true;
		self:StartMoving();
	end
end

local function OnDragStop(self)
	if self.isMoving then
		self.isMoving = false;
		self:StopMovingOrSizing();

		local x, y = self:GetLeft(), self:GetTop();
		
		local profile = AbsorbsViewer.db.profile;
		profile.titleX, profile.titleY = x, y;
	end
end

local function OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
	GameTooltip:SetText(L["ALT + Drag to move the frame anchor"], nil, nil, nil, nil, 1);
end

local function OnLeave(self)
	GameTooltip:Hide();
end

local backdrop = {bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeSize = 0.90,
		insets = {left = 1, right = 1, top = 1, bottom = 1}};
		
-- Note this function is exactly INVERSE to SortEffects in Core.lua
local function SortEffects(a, b)
	if a.priority == b.priority then
		return (a.amount > b.amount);
	else
		return (a.priority < b.priority);
	end
end

local function SetBorderVisibility(alpha)
	titleFrame:SetBackdropBorderColor(1, 1, 1, alpha and 0.6 or 0);
	
	for i, unit in pairs(activeUnitsList) do
		unit:SetBackdropBorderColor(1, 1, 1, alpha and 0.3 or 0);
	end
end

local function ResizeFrames()
	local scale = AbsorbsViewer.db.profile.frameScale;
	local width = AbsorbsViewer.db.profile.frameWidth;	
	
	baseFrame:SetScale(scale);
	titleFrame:SetWidth(width);
	
	local effectWidth = width - 14;
	
	for i, unit in pairs(activeUnitsList) do
		unit:SetWidth(width);
		
		for j, effect in pairs(unit.effectsList) do					
			effect:SetValue(1.0);
			effect:SetWidth(effectWidth);			
			effect.bg:SetWidth(effectWidth);
			
			if(effect.amount > 0) then
				effect:SetValue(effect.amount / effect.max);
			end
		end
	end
end

local function RepositionUnits()
	if(isHidden) then return; end

	for i, frame in pairs(activeUnitsList) do
		frame:ClearAllPoints();
		
		if(i > 1) then
			frame:SetPoint("TOPLEFT", activeUnitsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", titleFrame, "BOTTOMLEFT", 0, -2);
		end
	end
	
	lowestUnitFrame = activeUnitsList[#activeUnitsList];
end

local function RepositionEffects(guid)
	if(isHidden) then return; end

	if(not activeUnitsMap[guid]) then return; end
	
	local effectsList = activeUnitsMap[guid].effectsList;
	
	table.sort(effectsList, SortEffects);
	
	for i, frame in pairs(effectsList) do
		frame:ClearAllPoints();
	
		if(i > 1) then
			frame:SetPoint("TOPLEFT", effectsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", activeUnitsMap[guid], "TOPLEFT", 7, -22 );
		end
	end
end

local function ShowFrame()
	baseFrame:Show();
	
	RepositionUnits();
	
	for guid, unit in pairs(activeUnitsMap) do
		RepositionEffects(guid);
	end
	
	isHidden = false;
end

local function HideFrame()
	baseFrame:Hide();
	isHidden = true;
end

local function CreateBaseFrame()
	baseFrame = CreateFrame("Frame", "absorbsview_baseFrame", UIParent);
	baseFrame:SetScale(1.0);

	titleFrame = CreateFrame("Frame", "absorbsview_titleFrame", baseFrame);
	
	--titleFrame = baseFrame;
	
	titleFrame:SetHeight(20);
	titleFrame:SetMovable(true);
	titleFrame:EnableMouse(true);
	titleFrame:SetClampedToScreen(true);
	titleFrame:RegisterForDrag("LeftButton");
	titleFrame:SetBackdrop(backdrop);
	titleFrame:SetBackdropColor(0, 0, 0, 0.5);
	titleFrame:SetBackdropBorderColor(1, 1, 1, AbsorbsViewer.db.profile.hideBorders and 0.0 or 0.6);
	titleFrame:SetScript("OnDragStart", OnDragStart);
	titleFrame:SetScript("OnDragStop", OnDragStop);
	titleFrame:SetScript("OnShow", OnShow);
	titleFrame:SetScript("OnEnter", OnEnter);
	titleFrame:SetScript("OnLeave", OnLeave);
	titleFrame.name = "absorbs_viewer";
	
	titleFrame.text = titleFrame:CreateFontString(nil, "OVERLAY");
	titleFrame.text:SetPoint("LEFT", titleFrame, 10, 0);
	titleFrame.text:SetFontObject(GameFontHighlightSmall);
	titleFrame.text:SetText(L["Absorb Effects"]);
	
	titleFrame:SetScale(AbsorbsViewer.db.profile.frameScale);
	
	titleFrame:SetWidth(AbsorbsViewer.db.profile.frameWidth);	
	
	titleFrame:SetAlpha(1.0);
	
	titleFrame:ClearAllPoints();
	titleFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", AbsorbsViewer.db.profile.titleX, AbsorbsViewer.db.profile.titleY);
	
	baseFrame:ClearAllPoints();
	baseFrame:SetPoint("CENTER", titleFrame, "CENTER");
	
	titleFrame.CloseButton = CreateFrame("Button", nil, titleFrame);
	titleFrame.CloseButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp");
	titleFrame.CloseButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp");
	titleFrame.CloseButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");
	titleFrame.CloseButton:SetWidth(20);
	titleFrame.CloseButton:SetHeight(20);
	titleFrame.CloseButton:SetPoint("TOPRIGHT", titleFrame, "TOPRIGHT", 0, 0);
	titleFrame.CloseButton:SetScript("OnClick", HideFrame);
	
	lowestUnitFrame = titleFrame;
end

local function GetUnitFrame()
	if(#framePool_Units > 0) then
		return table.remove(framePool_Units, 1);
	end
	
	local frame = CreateFrame("Frame", nil, baseFrame);
	
	frame:SetBackdrop(backdrop);
	frame:SetBackdropColor(0, 0, 0, 0.5);
	frame:SetBackdropBorderColor(1, 1, 1, AbsorbsViewer.db.profile.hideBorders and 0.0 or 0.3);	
	
	local path, size, style = GameFontHighlight:GetFont();	
	
	frame.textName = frame:CreateFontString(nil, "OVERLAY");
	frame.textName:SetPoint("TOPLEFT", frame, 10, -5);
	frame.textName:SetFontObject(GameFontHighlightSmall);
	frame.textName:SetFont(path, size, style);
	
	frame.textTotal = frame:CreateFontString(nil, "OVERLAY");
	frame.textTotal:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.textTotal:SetFontObject(GameFontHighlightSmall);
	frame.textTotal:SetFont(path, size, style);
	
	frame:SetScale(1.0);
	
	frame:SetWidth(AbsorbsViewer.db.profile.frameWidth);	
	
	frame:SetAlpha(1.0);
	
	frame:ClearAllPoints();
	
	return frame;
end

local function RemoveUnit(guid)
	if(not activeUnitsMap[guid]) then return; end
	
	local unitFrame = activeUnitsMap[guid];
	
	activeUnitsMap[guid] = nil;
	
	for i, frame in pairs(activeUnitsList) do
		if(frame == unitFrame) then
			table.remove(activeUnitsList, i);
			break;
		end
	end
		
	unitFrame.effectsMap = nil;
	unitFrame.effectsList = nil;
	
	unitFrame:Hide();
	table.insert(framePool_Units, unitFrame);
	
	
	if(#activeUnitsList > 0) then
		RepositionUnits();
	else
		lowestUnitFrame = titleFrame;
	end
end

local function GetEffectFrame()
	if(#framePool_Effects > 0) then
		return table.remove(framePool_Effects, 1);
	end
		
	local frame = CreateFrame("StatusBar", nil, baseFrame);
	frame:Hide();		
	
	frame:SetClampedToScreen(true);
	frame:SetMinMaxValues(0, 1);
		
	frame.bg = CreateFrame("StatusBar", nil, frame);
	frame.bg:SetMinMaxValues(0, 1);
	frame.bg:SetValue(1.0);
	frame.bg:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0);
	frame.bg:SetFrameLevel(0);
	frame.bg:SetStatusBarTexture([[Interface\TargetingFrame\UI-StatusBar]]);
	frame.bg:SetWidth(AbsorbsViewer.db.profile.frameWidth - 14);
	frame.bg:SetHeight(14);
	frame.bg:SetStatusBarColor(0.5, 0.5, 0.5, 1.0);
	
	frame:SetStatusBarTexture([[Interface\TargetingFrame\UI-StatusBar]]);	
	frame:SetWidth(AbsorbsViewer.db.profile.frameWidth - 14);
	frame:SetHeight(14);
	frame:SetScale(1.0);
	
	local path, size, style = GameFontHighlight:GetFont();	
	size = size - 3;
	
	frame.textName = frame:CreateFontString(nil, "OVERLAY");
	frame.textName:SetFontObject(GameFontHighlight);
	frame.textName:SetJustifyH("LEFT");
	frame.textName:SetPoint("LEFT", frame, "LEFT", 0, 0);
	frame.textName:SetFont(path, size, style);	
			
	frame.textAmount = frame:CreateFontString(nil, "OVERLAY");
	frame.textAmount:SetFontObject(GameFontHighlight);
	frame.textAmount:SetJustifyH("RIGHT");
	frame.textAmount:SetPoint("RIGHT", frame, "RIGHT", -3, 0);
	frame.textAmount:SetFont(path, size, style);
		
	frame:SetAlpha(1.0);
	
	return frame;
end

local function ReleaseEffectFrame(frame)
	frame:Hide();
	
	table.insert(framePool_Effects, frame);
end

local function AddEffect(guid, unitName, spellId, amount, quality)
	local unitFrame, effectFrame;

	if(not activeUnitsMap[guid]) then
		unitFrame = GetUnitFrame();
		
		unitFrame:SetHeight(28);
		
		unitFrame.textName:SetText(unitName);
		unitFrame.textName:SetTextColor(GetUnitClassColorRGB(unitName));
		
		activeUnitsMap[guid] = unitFrame;
		table.insert(activeUnitsList, unitFrame);
		
		if(not isHidden) then
			unitFrame:SetPoint("TOPLEFT", lowestUnitFrame, "BOTTOMLEFT", 0, -2);
		end
		
		lowestUnitFrame = unitFrame;
		
		unitFrame:Show();
	else
		unitFrame = activeUnitsMap[guid];
	end
	
	effectFrame = GetEffectFrame();
	
	effectFrame.spellId = spellId;
	
	effectFrame:SetValue(1.0);
	effectFrame.amount = amount;
	effectFrame.max = amount;
	effectFrame.priority = AbsorbEffects[spellId][1];
	
	effectFrame:SetStatusBarColor(1 - quality, quality, 0, 0.6);
	
	if(SpellNames[spellId]) then
		effectFrame.textName:SetText(SpellNames[spellId]);
	else
		effectFrame.textName:SetText(L["Spell"].." "..spellId);
	end
	
	if(amount > 0) then
		effectFrame.textAmount:SetText(amount.." (100%)");
	end
	
	if(not unitFrame.effectsMap) then
		effectFrame:SetPoint("TOPLEFT", unitFrame, "TOPLEFT", 7, -22);
				
		unitFrame.effectsMap = {[spellId] = effectFrame};
		unitFrame.effectsList = {effectFrame};
	else		
		unitFrame.effectsMap[spellId] = effectFrame;			
		table.insert(unitFrame.effectsList, effectFrame);
		
		RepositionEffects(guid);
	end		
		
	unitFrame:SetHeight(unitFrame:GetHeight() + 16);
	effectFrame:Show();
end

local function RemoveEffect(guid, spellId)
	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local unitFrame = activeUnitsMap[guid];
	
		ReleaseEffectFrame(unitFrame.effectsMap[spellId]);
		
		unitFrame.effectsMap[spellId] = nil;
		
		for i, frame in pairs(unitFrame.effectsList) do
			if(frame.spellId == spellId) then
				table.remove(unitFrame.effectsList, i);
				break;
			end
		end
		
		if(#(unitFrame.effectsList) == 0) then			
			RemoveUnit(guid);
		else
			RepositionEffects(guid);
			unitFrame:SetHeight(unitFrame:GetHeight() - 16);
		end
	end		
end

local function UpdateEffect(guid, spellId, amount)
	if(amount == 0) then
		RemoveEffect(guid, spellId);
	end

	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local effectFrame = activeUnitsMap[guid].effectsMap[spellId];
		
		effectFrame.amount = amount;
		
		if(amount >= effectFrame.max) then
			effectFrame:SetValue(1.0);
			effectFrame.max = amount;
			
			-- We might need to reposition now
			RepositionEffects(guid);
		else
			effectFrame:SetValue(amount / effectFrame.max);	
		end
		
		effectFrame.textAmount:SetText(amount.." ("..string.format("%.0f", (amount * 100) / effectFrame.max).."%)");
	end
end

local function SetupEffectFilter()
	
end


-- #################################################################################################


local function getProfileOption(info)
	return AbsorbsViewer.db.profile[info.arg];
end

local function setProfileOption(info, value)
	AbsorbsViewer.db.profile[info.arg] = value;
end


-- #################################################################################################


local defaults = {
	profile = {
		titleX = 500,
		titleY = 500,
	
		show = true,
		frameScale = 1.0,
		frameWidth = 200,
		hideTitle = false,
		hideBorders = false,		
		
		filterAppliedByMe = false,
		filterAppliedByMe = false,
	},
};


-- #################################################################################################


local options = {
	type = "group",
	get = getProfileOption,
	set = setProfileOption,
	
	args = {
		general = {
			type = "group",
			name = L["General Options"],
			cmdInline = true,
			order = -1,
			
			args = {
				options = {
					type = "execute",
					order = 1,
					name = L["Options Window"],
					desc = L["Shows the Options Window"],
					func = function()
						AbsorbsViewer:ShowOptions();
					end,
					guiHidden = true,
				},			
				enabled = {
					type = "toggle",
					order = 2,
					name = L["Enabled"],
					desc = L["Enables/Disables this addon"],
					get = function() return AbsorbsViewer:IsEnabled() end,
					set = function()
						if(AbsorbsViewer:IsEnabled()) then
							AbsorbsViewer:Disable();
						else
							AbsorbsViewer:Enable()
						end
					end,
				},
				show = {
					type = "toggle",
					order = 3,
					name = L["Show"],
					desc = L["Displays the AbsorbsViewer window"],
					arg = "show",
					set = function(info, value)
						if(isHidden) then
							ShowFrame();
						else
							HideFrame();
						end
						
						setProfileOption(info, value);
					end,
				},
				windowHeader = {
					type = "header",
					order = 4,
					name = L["Window"],
				},
				scale = {
					type = "range",
					order = 5,
					softMin = 0.5,
					softMax = 2.0,
					step = 0.05,					
					name = L["Frame Scale"],
					desc = L["Scales all frames by a certain factor"],
					arg = "frameScale",
					set = function(...)
						setProfileOption(...);
						ResizeFrames();
					end,
				},
				width = {
					type = "range",
					order = 6,
					softMin = 100,
					softMax = 400,
					step = 5,
					name = L["Frame Width"],
					desc = L["Width of the base frame and all unit frames"],
					arg = "frameWidth",
					set = function(...)
						setProfileOption(...);
						ResizeFrames();
					end,					
				},
				title = {
					type = "toggle",
					order = 7,
					name = L["Hide window title"],
					desc = L["Hides the windows title required to move the frame"],
					arg = "hideTitle",
					set = function(info, value)
						if(value) then
							titleFrame:Hide();
						else
							titleFrame:Show();
						end
						
						setProfileOption(info, value);
					end,
				},			
				borders = {
					type = "toggle",
					order = 8,
					name = L["Hide window borders"],
					desc = L["Hides the window borders for all frames"],
					arg = "hideBorders",
					set = function(info, value)
						if(value) then
							SetBorderVisibility(false);
						else
							SetBorderVisibility(true);
						end
						
						setProfileOption(info, value);
					end,
				},
				filterHeader = {
					type = "header",
					order = 9,
					name = L["Filters"],
				},
				appliedbyme = {
					type = "toggle",
					order = 10,					
					name = L["Show only my effects"],
					desc = L["Shows only absorb effects applied by you"],
					arg = "filterAppliedByMe",
				},				
				appliedonme = {
					type = "toggle",
					order = 11,					
					name = L["Show only effects on me"],
					desc = L["Shows only absorb effects applied on you"],
					arg = "filterAppliedOnMe",
				},
			},
		},
	},
};


-- #################################################################################################


function AbsorbsViewer:Absorbs_EffectApplied(obj, sourceGUID, sourceName, destGUID, destName, spellId, amount, quality)
	local profile = self.db.profile;
	
	if((sourceGUID ~= playerGUID) and (profile.filterAppliedByMe)) then return; end
	if((destGUID ~= playerGUID) and (profile.filterAppliedOnMe)) then return; end

	AddEffect(destGUID, destName, spellId, amount, quality);
end

function AbsorbsViewer:Absorbs_EffectUpdated(obj, guid, spellId, amount)
	UpdateEffect(guid, spellId, amount);
end

function AbsorbsViewer:Absorbs_EffectRemoved(obj, guid, spellId)
	RemoveEffect(guid, spellId);
end

function AbsorbsViewer:Absorbs_UnitUpdated(obj, guid, amount, quality)
	if(activeUnitsMap[guid]) then
		local textTotal = activeUnitsMap[guid].textTotal;
		
		textTotal:SetText(amount);
		textTotal:SetTextColor(1 - quality, quality, 0);
	end
end

function AbsorbsViewer:Absorbs_UnitCleared(obj, guid)
	if(activeUnitsMap[guid]) then
		-- no need to worry about any data structures here
		for i, frame in pairs(activeUnitsMap[guid].effectsList) do
			ReleaseEffectFrame(frame);
		end
		
		RemoveUnit(guid);
	end
end

local areaSpellMapping = {};

-- I know you can have only one area effect per unit active in this implementation,
-- but more than one is impossible ingame anyway
function AbsorbsViewer:Absorbs_AreaCreated(obj, sourceGUID, sourceName, triggerGUID, spellId, value, quality)
	AddEffect(triggerGUID, L["Area Effects by"].." "..sourceName, spellId, value, quality);
	
	areaSpellMapping[triggerGUID] = spellId;
end

function AbsorbsViewer:Absorbs_AreaUpdated(obj, triggerGUID, value)
	UpdateEffect(triggerGUID, areaSpellMapping[triggerGUID], value);
end

function AbsorbsViewer:Absorbs_AreaCleared(obj, triggerGUID)
	RemoveEffect(triggerGUID, areaSpellMapping[triggerGUID]);
	
	areaSpellMapping[triggerGUID] = nil;
end

function AbsorbsViewer:ShowOptions()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.profile);
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.general);
end

function AbsorbsViewer:OnInitialize()
	self.db = AceDB:New("AbsorbsViewerDB", defaults);
	
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db);
	options.args.profile.order = -2;
	
	AceConfig:RegisterOptionsTable("AbsorbsViewer", options, {"av", "absorbsviewer"});
	
	self.optionsFrames = {}
	self.optionsFrames.general = AceConfigDialog:AddToBlizOptions("AbsorbsViewer", nil, nil, "general");
	self.optionsFrames.profile = AceConfigDialog:AddToBlizOptions("AbsorbsViewer", "Profiles", "AbsorbsViewer", "profile");
end

function AbsorbsViewer:OnEnable()
	if(not baseFrame) then
		CreateBaseFrame();
	end
	
	playerGUID = UnitGUID("player");
	
	if(not self.db.profile.show) then
		HideFrame();
	end
	
	if(self.db.profile.hideTitle) then
		titleFrame:Hide();
	end
	
	AbsorbEffects = AbsorbsMonitor.Effects;
	SpellNames = AbsorbsMonitor.SpellNames;
	
	self:RegisterMessage("Absorbs_EffectApplied");
	self:RegisterMessage("Absorbs_EffectUpdated");
	self:RegisterMessage("Absorbs_EffectRemoved");
	
	self:RegisterMessage("Absorbs_AreaCreated");
	self:RegisterMessage("Absorbs_AreaUpdated");
	self:RegisterMessage("Absorbs_AreaCleared");
	
	self:RegisterMessage("Absorbs_UnitUpdated");
	self:RegisterMessage("Absorbs_UnitCleared");
end

function AbsorbsViewer:OnDisable()
end
